/*
 * IRQ-handling code for TiVo devices, 7317 vs. 7318 abstractor
 *
 * Read..., Put all the ICKY ifdefy code here.
 *
 * Copyright (C) 2005 TiVo Inc. All Rights Reserved.
 */

#include <asm/bcm7315.h>

#ifdef CONFIG_TGC
#include <asm/bcm7318.h>
#endif

static void do_upg_irq(struct pt_regs *regs);
static void do_IRQ(int irq, struct pt_regs *regs);

#ifdef CONFIG_TIVO_DEVEL

int debug_irqs;

#define DEBUG_IRQS debug_irqs

#else
#define DEBUG_IRQS 0
#endif

#ifdef CONFIG_TGC /* 7318, should be CONFIG_CPU_BRCM7318 :( */

#define GIO_BYTES 6

#define GIO_RO(r)   (TGC_GIO_##r - __PERIPHREG(0))
#define MGIO_RO(r)  (TGC_GIOEXT_##r - __PERIPHREG(0))

#define ENET_TOP_IUDMA_ENET_MIPS_PCI_IRQ_STS (BCM_ENET_BASE_7xxx + 0x2418)
#define ENET_TOP_IUDMA_ENET_MIPS_PCI_IRQ_MSK (BCM_ENET_BASE_7xxx + 0x241c)

#else /* 7317 */
#define GIO_BYTES 9


#define GIO_RO(r)   (BCM_GIO_##r - __PERIPHREG(0))
#define GIOX_RO(r)  (BCM_GIOEXT_##r - __PERIPHREG(0))
#define MGIO_RO(r)  (BCM_MGIO_##r - __PERIPHREG(0))

#endif

static const u8 gio_reg_offsets[GIO_BYTES][3] = {
    { GIO_RO(MASKLSB),   GIO_RO(RSTLSB),   GIO_RO(STATLSB)   },
    { GIO_RO(MASKHSB),   GIO_RO(RSTHSB),   GIO_RO(STATHSB)   },
    { GIO_RO(MASK2MSB),  GIO_RO(RST2MSB),  GIO_RO(STAT2MSB)  },
    { GIO_RO(MASKMSB),   GIO_RO(RSTMSB),   GIO_RO(STATMSB)   },
    { MGIO_RO(MASKLSB),  MGIO_RO(RSTLSB),  MGIO_RO(STATLSB)  },
    { MGIO_RO(MASKHSB),  MGIO_RO(RSTHSB),  MGIO_RO(STATHSB)  },
#ifndef CONFIG_TGC
    { MGIO_RO(MASK2MSB), MGIO_RO(RST2MSB), MGIO_RO(STAT2MSB) },
    { MGIO_RO(MASKMSB),  MGIO_RO(RSTMSB),  MGIO_RO(STATMSB)  },
    { GIOX_RO(MASKLSB),  GIOX_RO(RSTLSB),  GIOX_RO(STATLSB)  },
#endif
};

#undef GIO_RO
#undef GIOX_RO
#undef MGIO_RO

extern volatile u8* const bcm_reg_base;

#define GIO_REG(byte, ireg)	(bcm_reg_base + gio_reg_offsets[byte][ireg])


#define GIO_MASK	0
#define GIO_RST		1
#define GIO_STAT	2

static inline void mask_irq(unsigned int irq)
{
	unsigned int status;

	if (irq < 1) {
		// Nothing for IRQ 0
	} else if (irq < 33) {
		*BCM_INTMASK &= ~(1 << (irq - 1));
#ifdef CONFIG_TGC
		if (irq == BCM_LINUX_CPU_ENET_IRQ){
		    *((volatile unsigned long *)
		      ENET_TOP_IUDMA_ENET_MIPS_PCI_IRQ_MSK) &= ~0x2;
		}
#endif
	} else if (irq < 37) {
		// Nothing for IRQs 33-36
	} else if (irq < 45) {
		*BCM_PM1_IRQEN_LO &= ~(1 << (irq - 37));
	} else if (irq < 53) {
		*BCM_PM1_IRQEN_HI &= ~(1 << (irq - 45));
	} else if (irq < 62) {
		status = read_32bit_cp0_register(CP0_STATUS);
		status &= ~(1 << (irq - 53 + 8));
		write_32bit_cp0_register(CP0_STATUS, status);
	} else {
#ifdef CONFIG_TGC
	    if (irq < 134) {
#endif
		int byte = (irq - BCM_LINUX_GIO_IRQ_BASE) / 8;
		int bit = (irq - BCM_LINUX_GIO_IRQ_BASE) & 7;
		*GIO_REG(byte, GIO_MASK) &= ~(1 << bit);
#ifdef CONFIG_TGC
	    } else {
		*BCM_EXTINT &= ~((1 << BCM_EXTIRQMASK_SHIFT) << (irq - 134));
	    }
#endif
	}

	barrier();
}

#ifdef CONFIG_TGC
#define GPIO_IRQ_BOUNDS(IRQNO) (IRQNO >= 62 && IRQNO <= 133)
#else
#define GPIO_IRQ_BOUNDS(IRQNO) (IRQNO >= 62)
#endif

static inline void mask_and_ack_irq(unsigned int irq)
{
        if (GPIO_IRQ_BOUNDS(irq)) {
		int byte = (irq - BCM_LINUX_GIO_IRQ_BASE) / 8;
		int bit = (irq - BCM_LINUX_GIO_IRQ_BASE) & 7;
		*GIO_REG(byte, GIO_MASK) &= ~(1 << bit);
		*GIO_REG(byte, GIO_RST) = 1 << bit;
		barrier();
		*GIO_REG(byte, GIO_RST) = 0;
		barrier();
	} else {
		return mask_irq(irq);
	}
}

static inline void ack_irq(unsigned int irq)
{
        if (GPIO_IRQ_BOUNDS(irq)) {
		int byte = (irq - BCM_LINUX_GIO_IRQ_BASE) / 8;
		int bit = (irq - BCM_LINUX_GIO_IRQ_BASE) & 7;
		*GIO_REG(byte, GIO_RST) = 1 << bit;
		barrier();
		*GIO_REG(byte, GIO_RST) = 0;
	}
	barrier();
}

static inline void unmask_irq(unsigned int irq)
{
	unsigned int status;

	if (irq < 1) {
		// Nothing for IRQ 0
	} else if (irq < 33) {
		*BCM_INTMASK |= 1 << (irq - 1);
#ifdef CONFIG_TGC
		if (irq == BCM_LINUX_CPU_ENET_IRQ) {
		        *((volatile unsigned long *)
			  ENET_TOP_IUDMA_ENET_MIPS_PCI_IRQ_MSK) |= 0x2;
		}
#endif
	} else if (irq < 37) {
		// Nothing for IRQs 33-36
	} else if (irq < 45) {
		*BCM_PM1_IRQEN_LO |= 1 << (irq - 37);
	} else if (irq < 53) {
		*BCM_PM1_IRQEN_HI |= 1 << (irq - 45);
	} else if (irq < 62) {
		status = read_32bit_cp0_register(CP0_STATUS);
		status |= 1 << (irq - 53 + 8);
		write_32bit_cp0_register(CP0_STATUS, status);
	} else {
#ifdef CONFIG_TGC
	    if (irq < 134) {
#endif
		int byte = (irq - BCM_LINUX_GIO_IRQ_BASE) / 8;
		int bit = (irq - BCM_LINUX_GIO_IRQ_BASE) & 7;
		*GIO_REG(byte, GIO_MASK) |= 1 << bit;
#ifdef CONFIG_TGC
	    } else {
		*BCM_EXTINT |= ((1 << BCM_EXTIRQMASK_SHIFT) << (irq - 134));
	    }
#endif
	}
	barrier();
}

#ifdef CONFIG_TGC
static inline int ethernet_int_check(int irq, struct pt_regs *regs) {
	if (irq == BCM_LINUX_CPU_ENET_IRQ - 1) {
		if (*((volatile unsigned long *)
		      ENET_TOP_IUDMA_ENET_MIPS_PCI_IRQ_STS) & 
		    *((volatile unsigned long *)
		      ENET_TOP_IUDMA_ENET_MIPS_PCI_IRQ_MSK) & 0x2 ) {	
			
			do_IRQ(BCM_LINUX_CPU_ENET_IRQ, regs);
		}			
		else					
			printk("unsolicited ENET int!!!\n");
		return 1;
	}
	return 0;
}
#else
#define ethernet_int_check(x,y) (0)
#endif

static inline void do_7315_irq(struct pt_regs *regs)
{
	int i;
	u32 intr_stat = *BCM_INTSTATUS & *BCM_INTMASK;

	for (i = 31; i >= 0; i--) {
		if (intr_stat & (1 << i)) {
			if (i == BCM_LINUX_UPG_IRQ - 1) {
				do_upg_irq(regs);
				return;
			}
			if(ethernet_int_check(i, regs))
			        return;
			do_IRQ(i + 1, regs);
			return;
		}
	}
	if (DEBUG_IRQS >= 1) {
		printk(KERN_WARNING "spurious BCM int! "
		       "BCM_INTSTATUS=0x%08x\n", *BCM_INTSTATUS);
	}
	atomic_inc(&irq_err_count);
}

#ifdef CONFIG_TGC
static inline void do_7318_extern_irq(struct pt_regs *regs)
{
	int i;
	u32 intr_stat = (*BCM_EXTINT >> BCM_EXTIRQSTAT_SHIFT) & 
				    (*BCM_EXTINT >> BCM_EXTIRQMASK_SHIFT);

	for (i = 1; i >= 0; i--) {
		if (intr_stat & (1 << i)) {
			do_IRQ(i + 134, regs);
			return;
		}
	}

	if (DEBUG_IRQS >= 1) {
		printk(KERN_WARNING "spurious BCM ext int! "
		       "BCM_EXTINT=0x%08x status %x, cause %x\n",
		       *BCM_EXTINT, (unsigned int)regs->cp0_status,
		       (unsigned int)regs->cp0_cause);
	}
	atomic_inc(&irq_err_count);
}
#endif


#ifdef CONFIG_TGC
#define DO_VECTOR_EXTERN_IRQ(irqno) \
			if (irqno == 3 || irqno == 4) { \
				do_7318_extern_irq(regs);  \
				return; \
			}
#define INIT_IRQ_EXTERN() \
    tivo_setup_irq(BCM_LINUX_EXTERNAL1_IRQ - 1, &cascade); \
    tivo_setup_irq(BCM_LINUX_EXTERNAL2_IRQ - 1, &cascade);
#else
#define DO_VECTOR_EXTERN_IRQ(irqno) 
#define INIT_IRQ_EXTERN()
#endif
