///////////////////////////////////////////////////////////////////////////
//
// File:  bcmenet.h
//
// Copyright (c) 2005, TiVo Inc.
//
///////////////////////////////////////////////////////////////////////////
#ifndef BCMENET_H
#define BCMENET_H   

#include <linux/kernel.h>

#include "bcmenet_regs.h"

#define __bytereg           (volatile unsigned char  * const)
#define __shortreg          (volatile unsigned short * const)
#define __longreg           (volatile unsigned long  * const)

#define ENET_REG_BASE       ( KSEG1ADDR(0x10080000) )
#define __preg(offset)      ( ENET_REG_BASE + (offset) )

typedef struct {
    volatile unsigned long              ctl;
    volatile unsigned long              irq_status;
    volatile unsigned long              irq_mask;
    volatile unsigned long              max_burst;
    volatile unsigned long              desc_ptr;
    volatile unsigned long              desc_offset;
}iudma_channel_t;

#define IUDMA_CHNL_DESC_OFFSET(channel) (__longreg(&((channel)->desc_offset)))
#define IUDMA_CHNL_DESC_PTR(channel)    (__longreg(&((channel)->desc_ptr)))
#define IUDMA_CHNL_MAX_BURST(channel)   (__longreg(&((channel)->max_burst)))
#define IUDMA_CHNL_IRQ_MSK(channel)     (__longreg(&((channel)->irq_mask)))
#define IUDMA_CHNL_IRQ_STS(channel)     (__longreg(&((channel)->irq_status)))
#define IUDMA_CHNL_CTL(channel)         (__longreg(&((channel)->ctl)))


typedef struct {
    volatile unsigned long              cfg;
    volatile unsigned long              fc_ch1_thresh_lo;
    volatile unsigned long              fc_ch1_thresh_hi;
    volatile unsigned long              fc_ch1_flow_alloc;
    volatile unsigned long              revision;
    volatile unsigned long              tstctl;
    volatile unsigned long              pci_irq_sts;
    volatile unsigned long              pci_irq_msk;
    volatile unsigned long              r5k_irq_sts;
    volatile unsigned long              r5k_irq_msk;
    volatile unsigned long              diag_ctl;
    volatile unsigned long              diag_rdbk;
    volatile unsigned long              reserved52[52];
    volatile iudma_channel_t            rx_dma_channel;
    volatile unsigned long              reserved26[26];
    volatile iudma_channel_t            tx_dma_channel;
}enet_dma_controller_t;

typedef struct {
    volatile unsigned long       length_and_status;
    volatile unsigned long       buffer;
}dma_desc_t;

#define DMA_DESC_PS_OV_SHIFT                            0
#define DMA_DESC_PS_OV_MASK                             0x00000001UL
#define DMA_DESC_PS_OV                                  0x00000001UL

#define DMA_DESC_PS_CRC_SHIFT                           1
#define DMA_DESC_PS_CRC_MASK                            0x00000002UL
#define DMA_DESC_PS_CRC                                 0x00000002UL

#define DMA_DESC_PS_RXERR_SHIFT                         2
#define DMA_DESC_PS_RXERR_MASK                          0x00000004UL
#define DMA_DESC_PS_RXERR                               0x00000004UL

#define DMA_DESC_PS_NO_SHIFT                            4
#define DMA_DESC_PS_NO_MASK                             0x00000008UL
#define DMA_DESC_PS_NO                                  0x00000008UL

#define DMA_DESC_PS_LG_SHIFT                            5
#define DMA_DESC_PS_LG_MASK                             0x00000010UL
#define DMA_DESC_PS_LG                                  0x00000010UL

#define DMA_DESC_PS_MULT_SHIFT                          6
#define DMA_DESC_PS_MULT_MASK                           0x00000020UL
#define DMA_DESC_PS_MULT                                0x00000020UL

#define DMA_DESC_PS_BRDCAST_SHIFT                       7
#define DMA_DESC_PS_BRDCAST_MASK                        0x00000040UL
#define DMA_DESC_PS_BRDCAST                             0x00000040UL

#define DMA_DESC_PS_MISS_SHIFT                          8
#define DMA_DESC_PS_MISS_MASK                           0x00000080UL
#define DMA_DESC_PS_MISS                                0x00000080UL

#define DMA_DESC_PS_APPEND_CRC_SHIFT                    9
#define DMA_DESC_PS_APPEND_CRC_MASK                     0x00000100UL
#define DMA_DESC_PS_APPEND_CRC                          0x00000100UL

#define DMA_DESC_PS_UNDERRUN_SHIFT                      10 
#define DMA_DESC_PS_UNDERRUN_MASK                       0x00000200UL
#define DMA_DESC_PS_UNDERRUN                            0x00000200UL

#define DMA_DESC_WRAP_SHIFT                             12 
#define DMA_DESC_WRAP_MASK                              0x00001000UL
#define DMA_DESC_WRAP                                   0x00001000UL

#define DMA_DESC_SOP_SHIFT                              13 
#define DMA_DESC_SOP_MASK                               0x00002000UL
#define DMA_DESC_SOP                                    0x00002000UL

#define DMA_DESC_EOP_SHIFT                              14 
#define DMA_DESC_EOP_MASK                               0x00004000UL
#define DMA_DESC_EOP                                    0x00004000UL

#define DMA_DESC_OWN_SHIFT                              15 
#define DMA_DESC_OWN_MASK                               0x00008000UL
#define DMA_DESC_OWN                                    0x00008000UL

#define DMA_DESC_BLENGTH_SHIFT                          16
#define DMA_DESC_BLENGTH_MASK                           0x0fff0000UL



typedef struct {
    volatile unsigned long      low; 
    volatile unsigned long      high;
}pm_data_t;


#define PM_HIGH_DATA_H_SHIFT                            0
#define PM_HIGH_DATA_H_MASK                             0x0000ffffUL

#define PM_HIGH_VALID_SHIFT                             16
#define PM_HIGH_VALID_MASK                              0x00010000UL
#define PM_HIGH_VALID                                   0x00010000UL


typedef struct {
    volatile unsigned long           tx_gd_octets;     /* 0x200 */
    volatile unsigned long           tx_gd_pkts;       /* 0x204 */
    volatile unsigned long           tx_all_oct;       /* 0x208 */
    volatile unsigned long           tx_all_pkts;      /* 0x20C */
    volatile unsigned long           tx_brdcast;       /* 0x210 */
    volatile unsigned long           tx_mult;          /* 0x214 */
    volatile unsigned long           tx_64;            /* 0x218 */
    volatile unsigned long           tx_65_127;        /* 0x21C */
    volatile unsigned long           tx_128_255;       /* 0x220 */
    volatile unsigned long           tx_256_511;       /* 0x224 */
    volatile unsigned long           tx_512_1023;      /* 0x228 */
    volatile unsigned long           tx_1024_max;      /* 0x22C */
    volatile unsigned long           tx_jab;           /* 0x230 */
    volatile unsigned long           tx_ovr;           /* 0x234 */
    volatile unsigned long           tx_frg;           /* 0x238 */
    volatile unsigned long           tx_underrun;      /* 0x23C */
    volatile unsigned long           tx_col;           /* 0x240 */
    volatile unsigned long           tx_1_col;         /* 0x244 */
    volatile unsigned long           tx_m_col;         /* 0x248 */
    volatile unsigned long           tx_ex_col;        /* 0x24C */
    volatile unsigned long           tx_late;          /* 0x250 */
    volatile unsigned long           tx_def;           /* 0x254 */
    volatile unsigned long           tx_crs;           /* 0x258 */
    volatile unsigned long           tx_paus;          /* 0x25C */
}enet_tx_stats_t;

typedef struct {
    volatile unsigned long           rx_gd_octets;     /* 0x280 */
    volatile unsigned long           rx_gd_pkts;       /* 0x284 */
    volatile unsigned long           rx_all_oct;       /* 0x288 */
    volatile unsigned long           rx_all_pkts;      /* 0x28C */
    volatile unsigned long           rx_brdcast;       /* 0x290 */
    volatile unsigned long           rx_mult;          /* 0x294 */
    volatile unsigned long           rx_64;            /* 0x298 */
    volatile unsigned long           rx_65_127;        /* 0x29C */
    volatile unsigned long           rx_128_255;       /* 0x2A0 */
    volatile unsigned long           rx_256_511;       /* 0x2A4 */
    volatile unsigned long           rx_512_1023;      /* 0x2A8 */
    volatile unsigned long           rx_1024_max;      /* 0x2AC */
    volatile unsigned long           rx_jab;           /* 0x2B0 */
    volatile unsigned long           rx_ovr;           /* 0x2B4 */
    volatile unsigned long           rx_frg;           /* 0x2B8 */
    volatile unsigned long           rx_drop;          /* 0x2BC */
    volatile unsigned long           rx_crc_align;     /* 0x2C0 */
    volatile unsigned long           rx_und;           /* 0x2C4 */
    volatile unsigned long           rx_crc;           /* 0x2C8 */
    volatile unsigned long           rx_align;         /* 0x2CC */
    volatile unsigned long           rx_sym;           /* 0x2D0 */
    volatile unsigned long           rx_paus;          /* 0x2D4 */
    volatile unsigned long           rx_cntrl;         /* 0x2D8 */
}enet_rx_stats_t;

#define MAX_TX_DESCS                 64
#define MAX_RX_DESCS                 64 /* 128 - MAX_TX_DESCS */
#define DMA_MAX_BURST_LENGTH         0x40
/* 7038 seems to be ignoring the lower 8 bits so to accomodate 1518 bytes
 * (0x5EE), we have to round it up to 0x600
 */
#define BCM_ENET_ALIGNMENT           0x100
#define ETH_MAX_MTU_SIZE             0x600
#define MAX_PM_REGS                  8

typedef struct {
    volatile unsigned long           rx_config;        /* 0x00 */
    volatile unsigned long           rx_length;        /* 0x04 */
    volatile unsigned long           tx_length;        /* 0x08 */
    volatile unsigned long           unused1;          /* 0x1C */
    volatile unsigned long           mii_sc;           /* 0x10 : unused */
    volatile unsigned long           mii_data;         /* 0x14 : unused */
    volatile unsigned long           interrupt_mask;   /* 0x18 */
    volatile unsigned long           interrupt_status; /* 0x1C */
    volatile unsigned long           unused3[3];       /* 0x20 - 0x2c */
    volatile unsigned long           control;          /* 0x2c */
    volatile unsigned long           xmit_control;     /* 0x30 */
    volatile unsigned long           xmit_watermark;   /* 0x34 */
    volatile unsigned long           mib_ctrl;         /* 0x38 : unused */
    volatile unsigned long           unused5[5];       /* 0x3C - 0x50 */
    volatile unsigned long           rx_test;          /* 0x50 : unused */
    volatile unsigned long           gpio;             /* 0x54 : unused */

    volatile pm_data_t               pm_data[MAX_PM_REGS];/* 0x58 - 0x94 */

    volatile unsigned long           unused90[90];     /* 0x098 - 0x200 */

    volatile enet_tx_stats_t         tx_stats;         /* 0x200 - 0x260 */

    volatile unsigned long           unused8[8];       /* 0x0260 - 0x280 */

    volatile enet_rx_stats_t         rx_stats;         /* 0x280 - 0x2DC */

    volatile unsigned long           unused1865[1865]; /* 0x2DC - 0x2000 */

    volatile dma_desc_t              rx_desc_table[MAX_RX_DESCS];
                                                       /* 0x2000-0x2200 */
    volatile dma_desc_t              tx_desc_table[MAX_TX_DESCS];
                                                       /* 0x2200-0x23FF */

    volatile unsigned long           dma_controller; 
}enet_bridge_t;

#define ENET_BRIDGE_BASE       (ENET_REG_BASE)
#define SUN_TOP_CONTROL_SW_RESET    __longreg(KSEG1ADDR(0x10404010))

#define ENET_BRIDGE                                                 \
    ((volatile enet_bridge_t * const)ENET_BRIDGE_BASE)

#define ENET_BRIDGE_RX_CONFIG                                       \
    (__longreg(&((ENET_BRIDGE)->rx_config)))

#define ENET_BRIDGE_RX_LENGTH                                       \
    (__longreg(&((ENET_BRIDGE)->rx_length)))

#define ENET_BRIDGE_TX_LENGTH                                       \
    (__longreg(&((ENET_BRIDGE)->tx_length)))

#define ENET_BRIDGE_INT_MASK                                        \
    (__longreg(&((ENET_BRIDGE)->interrupt_mask)))

#define ENET_BRIDGE_INT_STATUS                                      \
    (__longreg(&((ENET_BRIDGE)->interrupt_status)))

#define ENET_BRIDGE_CONTROL                                         \
    (__longreg(&((ENET_BRIDGE)->control)))

#define ENET_BRIDGE_XMIT_CONTROL                                    \
    (__longreg(&((ENET_BRIDGE)->xmit_control)))

#define ENET_BRIDGE_XMIT_WATERMARK                                  \
    (__longreg(&((ENET_BRIDGE)->xmit_watermark)))

#if 0
#define ENET_BRIDGE_PM_DATA(n)                                      \
    ((volatile pm_data_t *const)&((ENET_BRIDGE)->pm_data_##n))
#endif

#define ENET_BRIDGE_PM_DATA(n)                                      \
    ((volatile pm_data_t *const)&((ENET_BRIDGE)->pm_data[n]))

#define IS_PM_VALID(pmreg)                                          \
    ( ((pmreg->high) & PM_HIGH_VALID_MASK ))

#define ENET_BRIDGE_RX_DESC_TABLE                                   \
    ((volatile dma_desc_t *const)&((ENET_BRIDGE)->rx_desc_table[0]))

#define ENET_BRIDGE_TX_DESC_TABLE                                   \
    ((volatile dma_desc_t *const)&((ENET_BRIDGE)->tx_desc_table[0]))


/* DMA controller registers */
#define ENET_DMA_CONTROLLER                                         \
    ((volatile enet_dma_controller_t * const)&((ENET_BRIDGE)->dma_controller))

#define ENET_DMA_CONTROLLER_CONFIG                                  \
    (__longreg(&((ENET_DMA_CONTROLLER)->cfg)))

#define ENET_DMA_CONTROLLER_CH1_THRESH_LO                           \
    (__longreg(&((ENET_DMA_CONTROLLER)->fc_ch1_thresh_lo)))

#define ENET_DMA_CONTROLLER_CH1_THRESH_HI                           \
    (__longreg(&((ENET_DMA_CONTROLLER)->fc_ch1_thresh_hi)))

#define ENET_DMA_CONTROLLER_CH1_FLOW_ALLOC                          \
    (__longreg(&((ENET_DMA_CONTROLLER)->fc_ch1_flow_alloc)))

#define ENET_DMA_CONTROLLER_REVISION                                \
    (__longreg(&((ENET_DMA_CONTROLLER)->revision)))

#define ENET_DMA_CONTROLLER_R5K_IRQ_STS                             \
    (__longreg(&((ENET_DMA_CONTROLLER)->r5k_irq_sts)))

#define ENET_DMA_CONTROLLER_R5K_IRQ_MSK                             \
    (__longreg(&((ENET_DMA_CONTROLLER)->r5k_irq_msk)))

#define ENET_DMA_CONTROLLER_RX_CHANNEL                              \
    ((volatile iudma_channel_t * const)(&((ENET_DMA_CONTROLLER)->rx_dma_channel)))

#define ENET_DMA_CONTROLLER_TX_CHANNEL                              \
    ((volatile iudma_channel_t * const)(&((ENET_DMA_CONTROLLER)->tx_dma_channel)))


#endif /* BCMENET_H */
